@@ -1,8 +1,14 @@
 #include "defs.h"
-#include <linux/kexec.h>
 
+#include "xlat/kexec_load_flags.h"
 #include "xlat/kexec_arch_values.h"
-#include "xlat/kexec_flags.h"
+
+#ifndef KEXEC_ARCH_MASK
+# define KEXEC_ARCH_MASK 0xffff0000
+#endif
+#ifndef KEXEC_SEGMENT_MAX
+# define KEXEC_SEGMENT_MAX 16
+#endif
 
 static void
 print_kexec_segments(struct tcb *tcp, unsigned long addr, unsigned long len)
@@ -23,7 +29,12 @@ print_kexec_segments(struct tcb *tcp, unsigned long addr, unsigned long len)
 # define seg_memsz \
 	(current_wordsize == 4 ? (uint64_t) seg.seg32.memsz : seg.seg64.memsz)
 #else
-	struct kexec_segment seg;
+	struct {
+		void *buf;
+		size_t bufsz;
+		void *mem;
+		size_t memsz;
+	} seg;
 # define sizeof_seg sizeof(seg)
 # define seg_buf seg.buf
 # define seg_bufsz seg.bufsz
@@ -83,7 +94,7 @@ sys_kexec_load(struct tcb *tcp)
 	n &= ~KEXEC_ARCH_MASK;
 	if (n) {
 		tprints("|");
-		printflags(kexec_flags, n, "KEXEC_???");
+		printflags(kexec_load_flags, n, "KEXEC_???");
 	}
 
 	return 0;
